from attrs import frozen


@frozen
class Disconnected:
    pass


@frozen
class Connecting:
    url: str


@frozen
class Connected:
    url: str


@frozen
class NoPeer:
    pass


@frozen
class ConnectingPeer:
    pass  # are more details relevant?


@frozen
class ConnectedPeer:
    pass  # are more details relevant?


# XXX union types need python 3.10 or later
ConnectionStatus = Disconnected | Connecting | Connected
PeerConnection = NoPeer | ConnectingPeer | ConnectedPeer


@frozen
class WormholeStatus(object):
    """
    Represents the current status of a wormhole for use by the outside
    """

    # are we connected to the Mailbox Server
    mailbox_connection: ConnectionStatus = Disconnected()

    # XXX there isn't really this for pure "wormholes" -- only the
    # file-transfer v1 Transit or Dilation know about "connection to a
    # real peer"
    # we believe we have communication with our peer
    peer_connection: PeerConnection = NoPeer()

    # there's the notion of "we have a mailbox", separate from the
    # above; worth revealing?


@frozen
class DilationStatus(object):
    """
    Represents the current status of a Dilated wormhole
    """

    # current Dilation phase (ever increasing, aka "generation")
    phase: int = -1;

    # are we connected to the Mailbox Server
    # XXX should this (and peer_connection) be a WormholeStatus instead?
    mailbox: ConnectionStatus = Disconnected()

    # we believe we have communication with our peer
    peer_connection: PeerConnection = NoPeer()

    # there's the notion of "we have a mailbox", separate from the
    # above; worth revealing?

    # "hints"? "active_hint"?
    # "are we re-connecting" can be inferred from "mailbox" + "phase"


# worth having an Interface for "there is a new status"? it's just a
# callable that takes a WormholeStatus ...



import zope.interface
from twisted.trial import unittest
from attr import Attribute
from attr._make import NOTHING
from wormhole.util import provides


class IFoo(zope.interface.Interface):
    """
    An interface.
    """

    def f():
        """
        A function called f.
        """


def simple_attr(name):
    return Attribute(
        name=name,
        default=NOTHING,
        validator=None,
        repr=True,
        cmp=None,
        eq=True,
        hash=None,
        init=True,
        converter=None,
        kw_only=False,
        inherited=False,
    )


class TestProvides(unittest.TestCase):
    """
    Tests for `provides`.
    """

    def test_success(self):
        """
        Nothing happens if value provides requested interface.
        """

        @zope.interface.implementer(IFoo)
        class C(object):
            def f(self):
                pass

        v = provides(IFoo)
        v(None, simple_attr("x"), C())

    def test_fail(self):
        """
        Raises `TypeError` if interfaces isn't provided by value.
        """
        value = object()
        a = simple_attr("x")

        v = provides(IFoo)
        with self.assertRaises(TypeError):
            v(None, a, value)

    def test_repr(self):
        """
        Returned validator has a useful `__repr__`.
        """
        v = provides(IFoo)
        assert (
            "<provides validator for interface {interface!r}>".format(
                interface=IFoo
            )
        ) == repr(v)
